from .cca_base import _CCA_Base
from .gcca import GCCA, KGCCA
from .iterative import _Iterative, PLS_ALS, PMD, CCA_ALS, ElasticCCA, ParkhomenkoCCA, SCCA, SCCA_ADMM, SpanCCA, SWCCA
from .mcca import MCCA, KCCA
from .rcca import CCA, rCCA, PLS
from .tcca import TCCA, KTCCA
